/*

   befdeb.c
   Befunge Debugger - v1.0, Sept 1993, Chris Pressey

   usage :

   befdeb <befunge-source> <input-file> <output-file> <stack-file>

   compilation :

   Amiga : dcc befdeb.c -o befdeb
   Unix  : acc befdeb.c -o befdeb

 */

/********************************************************* #INCLUDE'S */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <time.h>

/********************************************************** #DEFINE'S */

#define LINEWIDTH  79		/* 0 .. 78 */
#define PAGEHEIGHT 22		/* 0 .. 21 */

#define cur pg[y * LINEWIDTH + x]
#define shrink(s) s[strlen(s)-1]=0

/********************************************************* STRUCTURES */

struct stack			/* stack structure, for values on stack */
  {
    signed long val;
    struct stack *next;
  }
 *head;				/* head of stack */

/*************************************************** GLOBAL VARIABLES */

char pg[LINEWIDTH * PAGEHEIGHT]; /* befunge 'page' of source */
int x = 0, y = 0;		 /* x and y of the PC */
int dx = 1, dy = 0;		 /* direction of the PC */
int stringmode = 0;		 /* flag : are we in string mode? */
int noin = 0;                    /* flag : no input file */

/********************************************************* PROTOTYPES */

void push (signed long val);
signed long pop ();

/******************************************************* MAIN PROGRAM */

main (argc, argv)
     int argc;
     char **argv;
{
  FILE *f;
  FILE *fi;
  FILE *fo;
  FILE *fs;

  srand (time (0));

  if (argc < 5)
    {
      printf ("USAGE : %s <befunge-source> <input-file> <output-file> <stack-file>\n", argv[0]);
      exit (0);
    }
  if (f = fopen (argv[1], "r"))			      /*** Input Phase */
    {
      int x = 0, y = 0;
      while (!feof (f))
	{
	  cur = fgetc (f);
	  if (cur == '\n')
	    {
	      cur = ' ';
	      x = 0;
	      y++;
	      if (y >= PAGEHEIGHT)
		break;
	    }
	  else
	    {
	      x++;
	      if (x >= LINEWIDTH)
		{
		  x = 0;
		  y++;
		  if (y >= PAGEHEIGHT)
		    break;
		}
	    }
	}
      fclose (f);
    }
  else
    {
      printf ("Error : couldn't open '%s' for input.\n", argv[1]);
      exit (0);
    }

  if (stricmp(argv[2], "nil"))
  {
    if (!(fi = fopen (argv[2], "r")))
      {
        printf ("Error : couldn't open '%s' for input.\n", argv[2]);
        exit (0);
      }
  } else
  {
    fi = NULL;
    noin = 1;
  }

  if (!(fo = fopen (argv[3], "w")))
    {
      printf ("Error : couldn't open '%s' for output.\n", argv[3]);
      exit (0);
    }

  if (!(fs = fopen (argv[4], "w")))
    {
      printf ("Error : couldn't open '%s' for output.\n", argv[4]);
      exit (0);
    }

  printf ("%c[1;1H", 27);
  printf ("%c[2J", 27);

  for(y = 0; y < PAGEHEIGHT; y++)
  {
    for(x = 0; x < LINEWIDTH; x++)
    {
      if (isprint(cur)) printf("%c", cur);
    }
    printf("\n");
  }

  x = y = 0;

  while (cur != '@')				/*** Intepreting Phase */
    {
      printf ("%c[%d;%dH", 27, y+1, x+1);
      fflush (stdout);

      if (stringmode && (cur != '"'))
	push (cur);
      else if (isdigit (cur))
	push (cur - '0');
      else
	switch (cur)
	   {
	   case '>':		/* PC Right */
	     dx = 1;
	     dy = 0;
	     break;
	   case '<':		/* PC Left */
	     dx = -1;
	     dy = 0;
	     break;
	   case '^':		/* PC Up */
	     dx = 0;
	     dy = -1;
	     break;
	   case 'v':		/* PC Down */
	     dx = 0;
	     dy = 1;
	     break;
	   case '|':		/* Vertical 'If' */
	     dx = 0;
	     if (pop ())
	       dy = -1;
	     else
	       dy = 1;
	     break;
	   case '_':		/* Horizontal 'If' */
	     dy = 0;
	     if (pop ())
	       dx = -1;
	     else
	       dx = 1;
	     break;
	   case '+':		/* Add */
	     push (pop () + pop ());
	     break;
	   case '-':		/* Subtract */
             {
               long a = pop();
               long b = pop();
               push(b - a);
             }
	     break;
	   case '*':		/* Multiply */
	     push (pop () * pop ());
	     break;
	   case '/':		/* Integer Divide */
	     {
	       signed long a = pop ();
	       signed long b = pop ();
	       push (b / a);
	     }
	     break;
	   case '%':		/* Modulo */
	     {
	       signed long a = pop ();
	       signed long b = pop ();
	       push (b % a);
	     }
	     break;
	   case '\\':		/* Swap */
	     {
	       signed long a = pop ();
	       signed long b = pop ();
	       push (a);
	       push (b);
	     }
	     break;
	   case '.':		/* Pop Out Integer */
             fprintf (fo, "%ld ", pop ());
	     fflush (fo);
	     break;
	   case ',':		/* Pop Out ASCII */
	     fprintf (fo, "%c", pop ());
	     fflush (fo);
	     break;
	   case '"':		/* Toggle String Mode */
	     stringmode = !stringmode;
	     break;
	   case ':':		/* Duplicate */
	     {
	       signed long a = pop ();
	       push (a);
	       push (a);
	     }
	     break;
	   case '!':		/* Negate */
          if (pop())
            push(0);
          else
            push(1);
	     break;
	   case '`':
	     {
	       signed long b = pop ();
	       signed long a = pop ();
	       push (a > b);
	     }
	     break;
	   case '#':		/* Bridge */
	     x += dx;
	     y += dy;
	     break;
	   case '$':		/* Pop and Discard */
	     pop ();
	     break;
	   case '?':		/* Random Redirect */
	     switch ((rand () / 32) % 4)
		{
		case 0:
		  dx = 1;
		  dy = 0;
		  break;
		case 1:
		  dx = -1;
		  dy = 0;
		  break;
		case 2:
		  dx = 0;
		  dy = -1;
		  break;
		case 3:
		  dx = 0;
		  dy = 1;
		  break;
		}
	     break;
	   case '&':		/* Input Integer */
	     {
	       signed long b;
	       if (!noin)
	       {
	         fscanf (fi, "%ld", &b);
	         push (b);
	       } else push (0);
	     }
	     break;
	   case '~':		/* Input ASCII */
	     {
	       char c;
               if (!noin) c = fgetc (fi); else c = 0;
	       push (c);
	     }
	     break;
	   case 'g':		/* Get Value */
	     {
	       signed long y = pop (), x = pop ();
	       push (cur);
	     }
	     break;
	   case 'p':		/* Put Value */
	     {
	       signed long y = pop (), x = pop ();
               printf ("%c[%d;%dH", 27, y+1, x+1);
	       cur = pop ();
	       if (isprint (cur)) printf ("%c", cur); else printf(".");
	     }
	     break;
	   default:
	     break;
	   }
      x += dx;
      y += dy;
      if (x < 0)
	x = LINEWIDTH;
      else
	x = x % LINEWIDTH;
      if (y < 0)
	y = PAGEHEIGHT;
      else
	y = y % PAGEHEIGHT;
      {
        struct stack *s;
        for (s = head; s; s = s->next)
          fprintf(fs, "%ld ", s->val);
        fprintf(fs, "\n");
        fflush(fs);
      }
    }
  exit (0);
  if (fi) fclose (fi);
  fclose (fo);
  fclose (fs);
  printf ("%c[22;1H", 27);
}

/*
 * pushes a value onto the stack.
 */
void push (signed long val)
{
  struct stack *s;
  s = (struct stack *) malloc (sizeof (struct stack));
  s->val = val;
  s->next = head;
  head = s;
}

/*
 * pops a value off the stack. returns 0 in case of underflow.
 */
signed long pop ()
{
  signed long v;
  struct stack *s = head;
  if (s)
    {
      v = head->val;
      head = head->next;
      free (s);
      return v;
    }
  else
    {
      return 0;
    }
}
